@@ -95,10 +95,14 @@ struct page_pipe_buf {
 	unsigned int pipe_size;	/* how many pages can be fit into pipe */
 	unsigned int pages_in;	/* how many pages are there */
 	unsigned int nr_segs;	/* how many iov-s are busy */
+#define PPB_LAZY (1 << 0)
+	unsigned int flags;
 	struct iovec *iov;	/* vaddr:len map */
 	struct list_head l;	/* links into page_pipe->bufs */
 };
 
+#define PP_HOLE_PARENT (1 << 0)
+
 struct page_pipe {
 	unsigned int nr_pipes;	/* how many page_pipe_bufs in there */
 	struct list_head bufs;	/* list of bufs */
@@ -111,6 +115,7 @@ struct page_pipe {
 	unsigned int nr_holes;	/* number of holes allocated */
 	unsigned int free_hole;	/* number of holes in use */
 	struct iovec *holes;	/* holes */
+	unsigned int *hole_flags;
 	unsigned flags;		/* PP_FOO flags below */
 };
 
@@ -120,10 +125,24 @@ struct page_pipe {
 
 struct page_pipe *create_page_pipe(unsigned int nr_segs, struct iovec *iovs, unsigned flags);
 extern void destroy_page_pipe(struct page_pipe *p);
-extern int page_pipe_add_page(struct page_pipe *p, unsigned long addr);
-extern int page_pipe_add_hole(struct page_pipe *p, unsigned long addr);
+extern int page_pipe_add_page(struct page_pipe *p, unsigned long addr,
+			      unsigned int flags);
+extern int page_pipe_add_hole(struct page_pipe *pp, unsigned long addr,
+			      unsigned int flags);
 
 extern void debug_show_page_pipe(struct page_pipe *pp);
 void page_pipe_reinit(struct page_pipe *pp);
 
+extern void page_pipe_destroy_ppb(struct page_pipe_buf *ppb);
+
+struct pipe_read_dest {
+	int p[2];
+	int sink_fd;
+};
+
+extern int pipe_read_dest_init(struct pipe_read_dest *prd);
+extern int page_pipe_read(struct page_pipe *pp, struct pipe_read_dest *prd,
+			  unsigned long addr, unsigned int *nr_pages,
+			  unsigned int ppb_flags);
+
 #endif /* __CR_PAGE_PIPE_H__ */
