@@ -101,6 +101,14 @@ struct thread_restore_args {
 
 typedef long (*thread_restore_fcall_t) (struct thread_restore_args *args);
 
+struct restore_vma_io {
+	int nr_iovs;
+	loff_t off;
+	struct iovec iovs[0];
+};
+
+#define RIO_SIZE(niovs)	(sizeof(struct restore_vma_io) + (niovs) * sizeof(struct iovec))
+
 struct task_restore_args {
 	struct thread_restore_args	*t;			/* thread group leader */
 
@@ -109,6 +117,9 @@ struct task_restore_args {
 	unsigned int			loglevel;
 	struct timeval			logstart;
 
+	int				uffd;
+	bool				has_thp_enabled;
+
 	/* threads restoration */
 	int				nr_threads;		/* number of threads */
 	thread_restore_fcall_t		clone_restore_fn;	/* helper address for clone() call */
@@ -121,6 +132,10 @@ struct task_restore_args {
 	VmaEntry			*vmas;
 	unsigned int			vmas_n;
 
+	int				vma_ios_fd;
+	struct restore_vma_io		*vma_ios;
+	unsigned int			vma_ios_n;
+
 	struct restore_posix_timer	*posix_timers;
 	unsigned int			posix_timers_n;
 
@@ -176,9 +191,10 @@ struct task_restore_args {
 
 	bool				compatible_mode;
 
+	bool				can_map_vdso;
 #ifdef CONFIG_VDSO
 	unsigned long			vdso_rt_size;
-	struct vdso_symtable		vdso_sym_rt;		/* runtime vdso symbols */
+	struct vdso_maps		vdso_maps_rt;		/* runtime vdso symbols */
 	unsigned long			vdso_rt_parked_at;	/* safe place to keep vdso */
 #endif
 	void				**breakpoint;
